"""This module provides test actions."""

# Import standard modules
from pathlib import Path
from sys import exit as sys_exit, stderr
from typing import cast
from unittest import defaultTestLoader

# Import third-party-modules
from batcave.fileutil import slurp
from batcave.sysutil import rmpath, syscmd, SysCmdRunner
from junitparser import JUnitXml
from xmlrunner import XMLTestRunner
from yaml import full_load as yaml_load

# Import project modules
from .utils import DEFAULT_ENCODING, VjerAction, VjerStep, helm


class TestStep(VjerStep):
    """This class provides test support."""

    def _test_runner(self, test_name: str) -> None:
        """Run tester."""
        SysCmdRunner(test_name, *(self.step_info.targets if self.step_info.targets else [self.project.name]), **self.step_info.options).run()

    def pre(self) -> None:
        """Prepare for testing on first run."""
        super().pre()
        if self.step_info.is_first_step:
            self.log_message('Preparing test results directory', True)
            if Path(self.project.test_results_dir).exists():
                self.log_message(f'Removing test results directory: {self.project.test_results_dir}')
                rmpath(self.project.test_results_dir)
            if not Path(self.project.test_results_dir).exists():
                self.log_message(f'Creating clean test results directory: {self.project.test_results_dir}')
                Path(self.project.test_results_dir).mkdir(parents=True)

    def test_docker(self) -> None:
        """Lint method for Docker dockerfiles."""
        dockerfile = slurp(self.dockerfile)
        syscmd('docker', 'run', '--interactive', '--rm', 'hadolint/hadolint', input_lines=dockerfile, ignore_stderr=True)
        if self.step_info.build_test_stage:
            self.docker_build(target=self.step_info.build_test_stage)

    def test_flake8(self) -> None:
        """Run python flake8 linter."""
        self._test_runner('flake8')

    def test_helm(self) -> None:
        """Lint method for Helm charts."""
        helm('dependency', 'build', self.helm_chart_root)
        helm('lint', self.helm_chart_root, **self.helm_args)
        with open(self.helm_chart_root / 'Chart.yaml', encoding=DEFAULT_ENCODING) as yaml_stream:
            helm_info = yaml_load(yaml_stream)
        if helm_info['type'] != 'library':
            helm('template', self.helm_chart_root, **self.helm_args)

    def test_mypy(self) -> None:
        """Run python mypy linter."""
        self._test_runner('mypy')

    def test_pylint(self) -> None:
        """Run python pylint linter."""
        self._test_runner('pylint')

    def test_python_unittest(self) -> None:
        """Runs the Python unittest module framework."""
        XMLTestRunner(output=str(self.project.test_results_dir), failfast=True, verbosity=2).run(defaultTestLoader.discover(self.project.project_root))
        for junit_results in self.project.test_results_dir.iterdir():
            test_suite = JUnitXml.fromfile(str(junit_results))
            if test_suite.errors or test_suite.failures:
                print('Unit tests failed', file=stderr)
                sys_exit(1)
            junit_results.rename(junit_results.parent / f'junit-{junit_results.name}')


def test() -> None:
    """This is the main entry point."""
    VjerAction('test', cast(VjerStep, TestStep)).execute()

# cSpell:ignore syscmd xmlrunner dockerfiles



"""This module provides release actions."""

# Import standard modules
from typing import cast

# Import third-party modules
from batcave.sysutil import SysCmdRunner
from twine.commands.upload import upload as twine_upload
from twine.settings import Settings

# Import project modules
from .utils import helm, VjerAction, VjerStep

bumpver_update = SysCmdRunner('bumpver', 'update', syscmd_args={'ignore_stderr': True}).run


class ReleaseStep(VjerStep):
    """Provide release support.

    Attributes:
        is_pre_release: Specifies that this is not a pre_release action.
    """
    is_pre_release = False

    def _execute(self) -> None:
        if self.step_info.release_only and self.is_pre_release:
            self.log_message('Skipping on pre_release')
            return
        if self.step_info.pre_release_only and not self.is_pre_release:
            self.log_message('Skipping on release')
            return
        super()._execute()

    def release_bumpver(self) -> None:
        """Perform a bumpver on release."""
        bumpver_update(**(self.step_info.args if self.step_info.args else {'tag': 'final', 'tag-commit': True}))

    def release_docker(self) -> None:
        """Perform a release of a Docker image by tagging."""
        self._docker_init()
        default_tags = [self.version_tag.lower()]
        if not self.is_pre_release:
            default_tags.append(f'{self.image_name}:latest'.lower())
        self.tag_images(self.image_tag, self.step_info.tags if self.step_info.tags else default_tags)

    def release_flit_build(self) -> None:
        """Run a Python flit build."""
        self.flit_build()

    def release_github(self) -> None:
        """Create a GitHub release."""
        SysCmdRunner('gh', 'release', 'create', f'v{self.version}', title=f'Release {self.version}', latest=True, generate_notes=True).run()

    def release_helm(self) -> None:
        """Perform a release of a Helm chart."""
        helm('push', self.helm_package, self.helm_repo.name, **self.helm_repo.push_args)

    def release_increment_release(self) -> None:
        """Increment the project release version."""
        if hasattr(self.project, 'version_service'):
            self.log_message('Incrementing version service not supported...skipping')
            return
        version_tuple = self.project.version.split('.')
        version_tuple[len(version_tuple) - 1] = str(int(version_tuple[-1]) + 1)
        new_version = '.'.join(version_tuple)
        use_branch = self.step_info.increment_branch if self.step_info.increment_branch else self.git_client.CI_COMMIT_REF_NAME
        self.project.version = new_version
        self.log_message(f'Incrementing release to {new_version} on branch {use_branch}')
        self.commit_files('Automated pipeline version update check-in [skip ci]', use_branch, self.config.filename, file_updater=self.config.write)

    def release_pypi(self) -> None:
        """Perform a release of a Python package to PyPI."""
        twine_upload(Settings(repository_name=('testpypi' if self.step_info.test_pypi else 'pypi'),
                              username=self.step_info.username, password=self.step_info.password,
                              non_interactive=True, disable_progress_bar=True),
                     [f'{self.project.artifacts_dir}/*'])

    def release_setuptools_build(self) -> None:
        """Run a Python setuptools build."""
        self.setuptools_build()

    def release_tag_source(self) -> None:
        """Tag the source in Git with a release tag."""
        self.tag_source(self.release.release_tag, f'Release {self.release.release_tag}')


def release() -> None:
    """This is the main entry point."""
    VjerAction('release', cast(VjerStep, ReleaseStep)).execute()

# cSpell:ignore syscmd testpypi



"""This module provides build actions."""

# Import standard modules
from os import getenv
from pathlib import Path
from shutil import copyfile
from tempfile import mkdtemp
from typing import cast, Optional

# Import third-party modules
from batcave.fileutil import pack
from batcave.lang import str_to_pythonval
from batcave.sysutil import rmpath
from docker.errors import BuildError as DockerBuildError
from build import ProjectBuilder

# Import project modules
from .utils import VJER_ENV, VjerAction, VjerStep, helm


class BuildStep(VjerStep):
    """This class provides build support.

        Build processing flow:
          pre:
              remove an old artifact directory
              create the artifact directory
              update the version files
          execute:
              run the inheriting build() method for the project type
          post:
              revert the version files
    """
    def pre(self) -> None:
        """This method is run at the start of the build."""
        super().pre()
        if self.step_info.is_first_step:
            self.log_message('Preparing artifact directory', True)
            if Path(self.project.artifacts_dir).exists():
                self.log_message(f'Removing stale artifact directory: {self.project.artifacts_dir}')
                rmpath(self.project.artifacts_dir)
            if not Path(self.project.artifacts_dir).exists():
                self.log_message(f'Creating clean artifact directory: {self.project.artifacts_dir}')
                Path(self.project.artifacts_dir).mkdir(parents=True)
        self.update_version_files()

    def post(self) -> None:
        """This method is run at the end of the build."""
        super().post()
        self.log_message('Build Completed Successfully', True)

    def always_post(self) -> None:
        """This method is always run at the end of the build."""
        super().always_post()
        self.update_version_files(reset=True)

    def create_archive(self, name: str, what: list, /, *, location: Optional[str] = None, arc_type: Optional[str] = None, use_tmpdir: bool = False) -> None:  # pylint: disable=too-many-arguments
        """Helper function to create an archive. When complete, the archive is copied to the project build artifact directory.

        Args:
            name: The archive file name.
            what: The files to include in the archive.
            location (optional, default=None): The location of the files to archive. If None, the current directory is used.
            arc_type (optional, default=None): The type of the archive. If None, is inferred from the file extension.
            use_tmpdir (optional, default=False): If True, create the archive in a temporary directory.

        Returns:
            Nothing.
        """
        package_dir = Path(mkdtemp()) if use_tmpdir else self.project.artifacts_dir
        package = package_dir / name
        try:
            self.log_message(f'Creating "{package}" archive from: {",".join(what)}', True)
            pack(package, what, item_location=location, archive_type=str(arc_type), ignore_empty=False)
            copyfile(package, self.project.artifacts_dir / name)
        finally:
            if use_tmpdir:
                rmpath(package_dir)

    def build_docker(self) -> None:
        """Run a Docker build."""
        push_image = str_to_pythonval(getenv('VJER_DOCKER_PUSH', str(not VJER_ENV == 'local')))
        self._docker_init(login=push_image)
        self.log_message(f'Building docker image: {self.image_tag}', True)
        build_args = {'VERSION': self.project.version,
                      'BUILD_VERSION': self.build.build_version} | self.step_info.build_args
        platform_arg = {'platform': platform} if (platform := getenv('DOCKER_DEFAULT_PLATFORM', '')) else {}
        try:
            log = self.docker_client.client.images.build(rm=True, pull=True, tag=self.image_tag,
                                                         dockerfile=(self.dockerfile),
                                                         buildargs=build_args.toDict(),
                                                         path=str(self.project.project_root),
                                                         **platform_arg)[1]
            error = None
        except DockerBuildError as err:
            error = err
            log = err.build_log
        for line in log:
            if ('stream' in line) and (line['stream'] != '\n'):
                self.log_message(line['stream'].strip())
        if error:
            raise error
        if push_image:
            self.log_message('Pushing image to registry', True)
            self.registry_client.get_image(self.image_tag).push()
            if self.step_info.tags:
                self.tag_images(self.image_tag, [f'{self.image_name}:{t}' for t in self.step_info.tags])

    def build_flit(self) -> None:
        """Run a Python flit build."""
        self.flit_build()

    def build_helm(self) -> None:
        """Build method for Helm charts."""
        helm('dependency', 'build', self.helm_chart_root)
        self.helm_build()

    def build_setuptools(self) -> None:
        """Run a Python setuptools build."""
        ProjectBuilder(self.project.project_root).build('sdist', self.project.artifacts_dir)
        ProjectBuilder(self.project.project_root).build('wheel', self.project.artifacts_dir)


def build() -> None:
    """This is the main entry point."""
    VjerAction('build', cast(VjerStep, BuildStep)).execute()

# cSpell:ignore pythonval buildargs



"""Vjer CI/CD module.

This module provides utilities for making it easier to write CI/CD automations.

"""

__all__ = ('__title__', '__summary__', '__uri__',
           '__version__', '__build_name__', '__build_date__',
           '__author__', '__email__',
           '__license__', '__copyright__')

__title__ = 'Vjer'
__summary__ = 'CI/CD Toolkit'
__uri__ = 'https://github.com/arisilon/vjer/'

__version__ = '39.0.0rc16'
__build_name__ = 'vjer_39.0.0rc16-14632297672'
__build_date__ = '2025-04-24 02:49:39.002253'

__author__ = 'Jeffery G. Smith'
__email__ = 'web@pobox.com'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2025 Jeffery G. Smith'



"""This module provides deployment rollback actions."""

# Import standard module
from typing import cast

# Import project modules
from .utils import VjerAction, VjerStep, helm


class RollbackStep(VjerStep):
    """This class provides rollback support."""

    def rollback_helm(self) -> None:
        """Rollback method for Helm charts."""
        chart_name = self.step_info.chart_name if self.step_info.chart_name else self.project.name.lower()
        helm('rollback',
             self.step_info.release_name.lower() if self.step_info.release_name else chart_name,
             wait=True, **self.helm_args)


def rollback() -> None:
    """This is the main entry point."""
    VjerAction('rollback', cast(VjerStep, RollbackStep)).execute()

# cSpell:ignore vjer



"""This program prints the tool information.

Attributes:
    PRODUCTS (list): This list of products on which to report.
"""

# Import standard modules
from re import compile as re_compile

# Import third-party modules
from batcave.lang import CommandResult
from batcave.sysutil import syscmd, CMDError
from dotmap import DotMap

PRODUCTS = [DotMap(name='Docker', regex=re_compile('Docker version (.+)')),
            DotMap(name='Google Cloud SDK', command='gcloud', regex=re_compile(r'Google Cloud SDK ([\d\.]+) ')),
            DotMap(name='Helm', args=['version'], regex=re_compile(r'Version:"v([\d\.]+)"'))]


def tool_reporter() -> dict:
    """Construct the tool report.

    Returns:
        A dictionary representing the report. There are three members in the dictionary:
            tool_versions: a dictionary of the tools with their versions.
            helm_plugins: a list of the helm plugins.
            helm_repos: a list of the helm repositories.
    """
    tool_info = DotMap(tool_versions={})
    for product in PRODUCTS:
        tool_info.tool_versions[product.name] = get_version(product)

    tool_info.helm_plugins = get_helm_info('plugin')
    tool_info.helm_repos = get_helm_info('repo')
    return tool_info.toDict()


def get_version(product: DotMap) -> str | list:
    """Determine the version for the specified product.

    Args:
        product: The product for which the version should be returned.

    Returns:
        The version of the specified product.
    """
    version_command = product.command if product.command else product.name.lower()
    version_args = product.args if product.args else ['--version']
    version_info: CommandResult = []
    try:
        version_info = syscmd(version_command, *version_args, ignore_stderr=True, append_stderr=True)
    except FileNotFoundError:
        pass
    except CMDError as err:
        if not (('not found' in str(err)) or ('not be found' in str(err)) or ('command could not be loaded' in str(err))):
            raise
    if product.raw:
        return version_info
    return version[1] if (version := product.regex.search(' '.join([line.strip() for line in version_info]))) else 'Not Found'


def get_helm_info(info_type: str) -> dict:
    """Return the requested Helm info.

    Args:
        info_type: The type of helm info to return.

    Returns:
        A dictionary of the requested info.
    """
    helm_info = {}
    try:
        for line in syscmd('helm', info_type, 'list', ignore_stderr=True):
            if line.startswith('NAME'):
                continue
            (name, url) = line.split()[0:2]
            helm_info[name] = url
    except FileNotFoundError:
        helm_info['Helm'] = 'not installed'
    except CMDError as err:
        if 'no repositories to show' not in ''.join(err.vars['err_lines']):
            raise
        helm_info['found'] = 'none'
    return helm_info if helm_info else {'found': 'none'}

# cSpell:ignore syscmd



#!/usr/bin/env python3
"""This is the bootstrap program for running the Vjer actions.
Since this program can be called before the requirements are install, it can only import standard modules.
"""

# Import standard modules
from importlib import import_module
import os
from os import getenv
from pathlib import Path
from platform import platform, system
from sys import exit as sys_exit, stderr, version as python_version
from tomllib import load as load_toml

# Import third-party modules
from batcave.commander import Argument, Commander
from batcave.fileutil import slurp
from batcave.sysutil import CMDError, SysCmdRunner
from batcave.version import AppVersion, VersionStyle
from dotmap import DotMap
from flit.install import Installer

# Import local modules
from . import __title__, __version__, __build_name__, __build_date__
from .utils import apt, apt_install, VJER_ENV, pip_install, ProjectConfig, ConfigurationError, PROJECT_CFG_FILE, VjerStep

ACTIONS = ['test', 'build', 'deploy', 'rollback', 'pre_release', 'release', 'freeze']


def main() -> None:
    """The main entrypoint."""
    version = AppVersion(__title__, __version__, __build_date__, __build_name__)
    args = Commander('Vjer CI/CD Automation Tool', [Argument('actions', choices=ACTIONS, nargs='+')], version=version).parse_args()
    VjerStep().log_message(version.get_info(VersionStyle.one_line), True)
    _setup_environment()
    VjerStep().log_message(f'OS: {platform()}')
    if (system() == 'Linux') and (release_file := Path('/etc/os-release')).exists():
        for line in slurp(release_file):
            VjerStep().log_message(line.strip())
    VjerStep().log_message(f'Python version: {python_version}')

    _sys_initialize()
    for action in args.actions:
        action_module = import_module(f'vjer.{action}')
        action_module.__dict__[action]()


def _pip_setup() -> None:
    SysCmdRunner('python', '-m', 'pip', 'install', 'pip', quiet=True, no_cache_dir=True, upgrade=True).run()
    pip_install('setuptools', 'wheel')


def _toml_to_dotmap(file_path: str) -> DotMap:
    """Reads a TOML file and returns its content as a dictionary."""
    with open(file_path, 'rb') as file:
        return DotMap(load_toml(file))


def _setup_environment() -> None:
    try:
        config = ProjectConfig()
    except ConfigurationError as err:
        if err.code != ConfigurationError.CONFIG_FILE_NOT_FOUND.code:
            raise
        print('ERROR The Vjer configuration file was not found:', PROJECT_CFG_FILE, file=stderr)
        sys_exit(1)
    if (VJER_ENV == 'local') and not getenv('VIRTUAL_ENV', ''):
        print('ERROR Vjer must be run from a virtual environment.', file=stderr)
        sys_exit(1)
    if hasattr((config := ProjectConfig()).project, 'environment'):
        for (var, val) in config.project.environment.items():
            VjerStep().log_message(f'setting {var}={val}')
            os.environ[var] = val  # putenv doesn't work because the values are needed for this process.


def _sys_initialize() -> None:
    if pkg_installs := getenv('VJER_PKG_INSTALLS', ''):
        apt('update')
        apt_install(pkg_installs)

    if (pip_installs := getenv('VJER_PIP_INSTALLS', '')) or (pip_file := getenv('VJER_PIP_INSTALL_FILE', '')) or (pyproject_build := getenv('VJER_PYPROJECT_BUILD', '')):
        _pip_setup()
        if pip_installs:
            pip_install(pip_installs)
        if pip_file:
            pip_install(requirement=pip_file)
        if pyproject_build:
            pyproject = _toml_to_dotmap('pyproject.toml')
            if pyproject['build-system']['build-backend'] == 'setuptools.build_meta':
                try:
                    pip_install('.[test]')
                except CMDError as err:
                    if "WARNING: Failed to remove contents in a temporary directory" not in str(err):
                        raise
            if pyproject['build-system']['build-backend'] == 'flit.buildapi':
                Installer.from_ini_path(Path('pyproject.toml')).install()


if __name__ == '__main__':
    main()

# cSpell:ignore putenv buildapi



"""This module provides pre_release actions."""

# Import standard modules
from typing import cast, override

# Import project modules
from .release import ReleaseStep
from .utils import sanitize_tag, VjerAction, VjerStep


class PreReleaseStep(ReleaseStep):
    """Provide pre_release support.

    Attributes:
        is_pre_release: Specifies to the ReleaseStep parent class that this is a pre_release action.
    """
    is_pre_release = True

    def __init__(self):
        """Sets the project version to a pre_release value."""
        super().__init__()
        if self.project.version_service.type != 'bumpver':
            self.project.version = f'{self.project.version}-{self.pre_release_num}'

    @override
    def release_bumpver(self) -> None:
        """Perform a bumpver on release."""
        if not self.step_info.args:
            self.step_info.args = {'tag-num': True}
        super().release_bumpver()

    @override
    def release_helm(self) -> None:
        """Pre_release a Helm chart."""
        self._release_helm()
        for version in self.step_info.extra_versions:
            self.project.version = sanitize_tag(version).lower()
            self._release_helm()

    def _release_helm(self) -> None:
        if self.helm_repo.type == 'oci':
            self.update_version_files()
            try:
                self.helm_build()
            finally:
                self.update_version_files(reset=True)
        else:
            list(self.project.artifacts_dir.glob('*.tgz'))[0].rename(self.helm_package)
        super().release_helm()


def pre_release() -> None:
    """This is the main entry point."""
    VjerAction('release', cast(VjerStep, PreReleaseStep)).execute()



"""This module provides deployment actions."""

# Import standard module
from typing import cast

# Import project modules
from .utils import helm, VjerAction, VjerStep


class DeployStep(VjerStep):
    """This class provides deployment support."""

    def deploy_helm(self) -> None:
        """Deploy method for Helm charts."""
        chart_name = self.step_info.chart_name if self.step_info.chart_name else self.project.name.lower()
        release_name = self.step_info.release_name.lower() if self.step_info.release_name else chart_name
        helm_args = self.helm_args
        is_remote = self.step_info.remote is not False
        if is_remote:
            helm_chart = f'{self.helm_repo.name}/{chart_name}'
            if 'version' not in helm_args:
                helm_args['version'] = self.project.version
        else:
            helm_chart = self.helm_package
        if is_remote:
            helm('repo', 'update')
        helm('upgrade', release_name, helm_chart, install=True, atomic=True, wait=True, **helm_args)


def deploy() -> None:
    """This is the main entry point."""
    VjerAction('deploy', cast(VjerStep, DeployStep)).execute()



#!/usr/bin/env python3
"""This module creates a Python freeze file for the project.

Attributes:
    PROJECT_ROOT (Path): The root directory of the project.
    FREEZE_FILE_EXT (str): The extension for the frozen requirements file.
    REQUIREMENTS_FILE (Path): The full path of the frozen requirements file.
    FREEZE_FILE_BASE (str): The base name of the frozen requirements file.
    LINUX_MODULES (list): A list of modules that should be limited to the linux platform.
    WINDOWS_MODULES (list): A list of modules that should be limited to the Windows platform.
"""

# Import standard modules
import os
from pathlib import Path
from sys import argv
from tempfile import mkdtemp
from venv import EnvBuilder

# Import BatCave modules
from batcave.fileutil import slurp, spew
from batcave.lang import WIN32
from batcave.sysutil import rmpath, syscmd, SysCmdRunner

from .utils import DEFAULT_ENCODING

PROJECT_ROOT = Path.cwd()
FREEZE_FILE_EXT = '.txt'
REQUIREMENTS_FILE = (PROJECT_ROOT / 'requirements').with_suffix(FREEZE_FILE_EXT)
FREEZE_FILE_BASE = 'requirements-frozen'
LINUX_MODULES: list = []
WINDOWS_MODULES = ['WMI']

pip = SysCmdRunner('pip', show_cmd=False, show_stdout=False, syscmd_args={'ignore_stderr': True}).run


def freeze() -> None:
    """Create the requirement-freeze.txt file leaving out the development tools and adding platform specifiers."""
    freeze_file = (PROJECT_ROOT / (FREEZE_FILE_BASE + (f'-{argv[2]}' if (len(argv) > 2) else ''))).with_suffix(FREEZE_FILE_EXT)
    print('Creating virtual environment in:', venv_dir := Path(mkdtemp()))
    EnvBuilder(with_pip=True).create(venv_dir)
    python = ((venv_bin := venv_dir / ('Scripts' if WIN32 else 'bin')) / 'python').with_suffix('.exe' if WIN32 else '')
    os.environ['PATH'] = os.path.pathsep.join((str(venv_bin), os.environ['PATH']))
    print('Upgrading pip')
    syscmd(str(python), '-m', 'pip', 'install', '-qqq', '--upgrade', 'pip')
    print('Updating pip install tools')
    pip('install', '-qqq', 'setuptools', 'wheel', upgrade=True)
    print('Installing modules from', REQUIREMENTS_FILE)
    pip('install', '-qqq', upgrade=True, requirement=REQUIREMENTS_FILE)
    print('Creating frozen requirements file:', freeze_file)
    spew(freeze_file, pip('freeze', requirement=REQUIREMENTS_FILE))
    freeze_file_contents = [line.strip() for line in slurp(freeze_file)]
    with open(freeze_file, 'w', encoding=DEFAULT_ENCODING) as freeze_file_stream:
        for line in freeze_file_contents:
            module = line.split('==')[0]
            if ('win32' in module) or (module in WINDOWS_MODULES):
                line += "; sys_platform == 'win32'"
            if ('ansible' in module) or (module in LINUX_MODULES):
                line += "; sys_platform != 'win32'"
            print(line, file=freeze_file_stream)
    rmpath(venv_dir)

# cSpell:ignore syscmd



"""Module to hold common values and utility functions for CI/CD support scripts.

Attributes:
    PROJECT_CFG_FILE (str): The name of the project config file.
    TOOL_REPORT (Path): The path of the tool report.

    There are several tool runners defined for simplified usage: git, helm.
"""
# Import standard modules
from copy import deepcopy as copy_object
from datetime import datetime
from os import getenv
from pathlib import Path
from random import randint
from re import compile as re_compile, sub as re_sub
from shutil import copy, copyfile, copytree
from stat import S_IWUSR
from string import Template
from sys import exit as sys_exit, stderr
from typing import Callable, cast, Optional, override

# Import third-party modules
from batcave.automation import Action
from batcave.cloudmgr import Cloud, CloudType
from batcave.cms import Client, ClientType
from batcave.expander import Expander, file_expander
from batcave.lang import BatCaveError, BatCaveException, PathName, DEFAULT_ENCODING, WIN32, yaml_to_dotmap
from batcave.platarch import Platform
from batcave.sysutil import CMDError, SysCmdRunner, syscmd
from bumpver.config import init as bumpver_config
from dotmap import DotMap
from flit.build import main as flit_builder
from yaml import safe_dump as yaml_dump, safe_load as yaml_load

# Import project modules
from .tool_reporter import tool_reporter

_CONFIG_SECTIONS = ('project', 'test', 'build', 'deploy', 'rollback', 'release')
_PROJECT_DEFAULTS = DotMap(build_artifacts='artifacts',
                           build_num_var='VJER_BUILD_NUM',
                           chart_root='helm-chart',
                           container_registry=DotMap(type='local', auth=tuple(), name=''),
                           dockerfile='Dockerfile',
                           test_results='test_results',
                           version_service=DotMap(type='vjer'))
_VALID_SCHEMAS = [3]

PROJECT_CFG_FILE = getenv('VJER_CFG', 'vjer.yml')
TOOL_REPORT = Path(__file__).parent.absolute() / 'tool_report.yml'

HELM_CHART_FILE = 'Chart.yaml'
DEFAULT_VERSION_FILES = {'helm': [HELM_CHART_FILE, 'values.yaml'],
                         'flit': ['__init__.py'],
                         'setuptools': ['__init__.py']}

VJER_ENV = getenv('VJER_ENV', 'local')
REMOTE_REF = 'vjer_origin'

BAD_TAG_CHARACTERS = '? ,.:`"/\\\'*'
REPLACER_HOLDER = '-'

apt = SysCmdRunner('apt-get', '-y').run
apt_install = SysCmdRunner('apt-get', '-y', 'install', no_install_recommends=True).run
git = SysCmdRunner('git').run
helm = SysCmdRunner('helm', syscmd_args={'ignore_stderr': True}).run
pip_install = SysCmdRunner('pip', 'install', quiet=True, no_cache_dir=True, upgrade=True).run


class ConfigurationError(BatCaveException):
    """Configuration errors.

    Attributes:
        BAD_FORMAT: The format of the configuration file is bad.
        CONFIG_FILE_NOT_FOUND: The config file was not found.
        INVALID_SCHEMA: The schema of the configuration file is not supported.
    """
    BAD_FORMAT = BatCaveError(1, Template('Invalid format for configuration file: $file'))
    CONFIG_FILE_NOT_FOUND = BatCaveError(2, Template('Configuration file not found: $file'))
    INVALID_SCHEMA = BatCaveError(3, Template('Invalid configuration schema: found=$found, expected=$expected'))


class StepError(BatCaveException):
    """Step errors.

    Attributes:
        UNKNOWN_OBJECT: The specified object is of an unknown type.
    """
    UNKNOWN_OBJECT = BatCaveError(1, Template('Unknown $type: $name'))


class Environment:  # pylint: disable=too-few-public-methods
    """Provides an interface to the environment variables."""

    def __getattr__(self, attr: str):
        value = getenv(attr)
        if value is None:
            raise AttributeError(f'Environment variable not found: {attr}')
        return value


class GitClient(Environment):
    """Provides an interface to the Git server environment and API."""

    def __init__(self, project_id: Optional[str] = None, client_root: Optional[PathName] = None, branch: Optional[str] = None):
        """
        Args:
            project_id (optional, default=None): The ID of the Git project.
            client_root (optional, default=None): The root directory of the project.
            branch (optional, default=None): The branch to use for the Git client.

        Attributes:
            CI_REMOTE_REF: The full git reference to the project.
            branch: The value of the branch argument.
            client: The git client for the project.
            project_id: The value of the project_id argument.
        """
        self.project_id = project_id
        self.client = Client(ClientType.git, 'vjer', connect_info=str(client_root), create=False) if (client_root and (Path(client_root) / '.git').exists()) else None
        self.branch = branch if branch else getattr(self, 'CI_COMMIT_BRANCH', '')

    def __enter__(self):
        return self

    def __exit__(self, *_unused_exc_info):
        return False


class ConfigSection:
    """Base class to manage configuration sections."""

    def __init__(self, **defaults):
        """
        Args:
            **defaults: The default items for the configuration.

        Attributes:
            _default_property_holders: A list of property holders for defaults.
            _defaults: The configuration defaults.
            _expander: The expander to use for variable replacement.
            _values: The configuration values.
        """
        self._values = DotMap()
        self._defaults = DotMap(**defaults)
        self._default_property_holders = [Environment()]
        self._expander = None
        self.update_expander(property_holders=self._default_property_holders)

    def __getattr__(self, attr: str):
        for config in (self._values, self._defaults):
            if attr in config:
                value = getattr(config, attr)
                if isinstance(value, list):
                    return [self._expander.expand(v) for v in value]
                if isinstance(value, dict):
                    return DotMap({k: self._expander.expand(v) for (k, v) in value.items()})
                if isinstance(value, str):
                    return self._expander.expand(value)
                return value
        raise AttributeError(f'No configuration value found: {attr}')

    def __setattr__(self, attr: str, value: str):
        if attr.startswith('_'):
            super().__setattr__(attr, value)
            return
        setattr(self._values, attr, value)

    values = property(lambda s: s._values.toDict(), doc='A read-only property which returns the configuration values.')

    def update_expander(self, *, property_holders: Optional[list] = None, property_dict: Optional[dict] = None) -> None:
        """Set the expander property holders.

        Args:
            property_holders (optional, default=None): A list of property holders from which to read values.
            property_dict (optional, default=None): A dictionary from which to read values.

        Returns:
            Nothing.
        """
        if property_holders:
            self._expander = Expander(var_props=property_holders + self._default_property_holders)
        if property_dict:
            self._expander.var_dict |= property_dict

    def update(self, values: dict | DotMap, /) -> None:
        """Update the configuration section values.

        Args:
            values: Update the configuration values from the provided dictionary.

        Returns:
            Nothing.
        """
        self._values |= values

    def update_defaults(self, values: dict | DotMap, /) -> None:
        """Updates the configuration section default values.

        Args:
            values: Update the defaults with the provided dictionary.

        Returns:
            Nothing.
        """
        self._defaults |= values


class ProjectConfig:
    """Stores project related configuration items."""

    def __init__(self):
        """
        Attributes:
            _config_file: The config file for the project configuration.
            _sections: A list of the sections in the project configuration.
            schema: The schema version of the project configuration.
        """
        project_root = Path.cwd()
        self._sections = {'project': ConfigSection(**(DotMap(project_root=project_root) | _PROJECT_DEFAULTS)),
                          'test': ConfigSection(),
                          'deploy': ConfigSection(clean=True),
                          'rollback': ConfigSection(clean=True),
                          'release': ConfigSection()}
        self._sections['build'] = ConfigSection(source_dir=self.project.project_root / 'src',
                                                version_files=[],
                                                artifacts={},
                                                build_date=str(datetime.now()),
                                                platform=Platform().bart)
        self._config_file = self.project.project_root / PROJECT_CFG_FILE
        self._load_config()
        self._set_defaults()
        self._set_version()

        for section in _CONFIG_SECTIONS:
            self._sections[section].update_expander(property_holders=list(self._sections.values()))

    def __getattr__(self, attr: str):
        if attr not in self._sections:
            raise AttributeError(f'Configuration section not found: {attr}')
        return self._sections[attr]

    def _get_phase_step(self, phase: str, step_type: str) -> DotMap:
        """Get the specified step type for the specified phase.

        Args:
            phase: The phase for which to return the step.
            step_type: The step type to return from the phase.

        Returns:
            A DotMap for the step.
        """
        if hasattr(phase_ref := getattr(self, phase), 'steps'):
            for step in phase_ref.steps:
                if step.get('type') == step_type:
                    return copy_object(step)
        return DotMap(type=step_type)

    def _load_config(self) -> None:
        if not self._config_file.exists():
            raise ConfigurationError(ConfigurationError.CONFIG_FILE_NOT_FOUND, file=self._config_file)
        yaml_as_dict = DotMap(schema=0)
        with open(self._config_file, encoding=DEFAULT_ENCODING) as config_file:
            yaml_as_dict |= DotMap(yaml_load(config_file))
        if not yaml_as_dict:
            raise ConfigurationError(ConfigurationError.BAD_FORMAT, file=self._config_file)
        if yaml_as_dict.schema not in _VALID_SCHEMAS:
            raise ConfigurationError(ConfigurationError.INVALID_SCHEMA, found=yaml_as_dict.schema, expected=_VALID_SCHEMAS)
        self.schema = yaml_as_dict.schema
        for section in _CONFIG_SECTIONS:
            if section in yaml_as_dict:
                self._sections[section].update(yaml_as_dict[section])

    def _set_defaults(self) -> None:
        self.project.update_defaults(DotMap(artifacts_dir=self.project.project_root / self.project.build_artifacts,
                                            test_results_dir=self.project.project_root / self.project.test_results))

    def _set_version(self) -> None:
        match self.project.version_service.type:
            case 'bumpver':
                self.project.version = bumpver_config()[1].current_version
            case 'environment':
                self.project.version = getenv(self.project.version_service.variable, '').rstrip('.')
            case 'vjer':
                pass
            case _:
                print('Unknown version service:', self.project.version_service.type, file=stderr)
                sys_exit(1)
        build_num = getenv(self.project.build_num_var, '0')
        build_version = f'{self.project.version}-{build_num}'
        self.build.update_defaults(DotMap(build_num=build_num,
                                          build_version=build_version,
                                          build_version_msbuild=f'{self.project.version}.{build_num}',
                                          build_name=f'{self.project.name}_{build_version}'))
        self.release.update_defaults(DotMap(release_tag=f'v{self.project.version}'))
        for (piece, index) in {'major': 0, 'minor': 1, 'patch': 2}.items():
            self.project.update_defaults({f'{piece}': self.project.version.split('.', 2)[index]})

    filename = property(lambda s: s._config_file, doc='A read-only property which returns the configuration file name.')

    def write(self) -> None:
        """Writes out the project configuration.

        Returns:
            Nothing.
        """
        with open(self._config_file, 'w', encoding=DEFAULT_ENCODING) as config_file:
            yaml_dump({'schema': self.schema} | {s: c.values for (s, c) in self._sections.items() if c.values},
                      config_file, indent=2)


class VjerStep(Action):  # pylint: disable=too-many-instance-attributes
    """ Class to represent a single Action step."""

    def __init__(self):
        """
        Attributes:
            build: The project build configuration.
            config: The project configuration.
            docker_client: The Docker client.
            git_client: The Git repository for the project.
            image_name: The Docker image name.
            image_tag: The Docker image tag.
            pre_release_num: The pre_release suffix for the project.
            project: The project name.
            registry_client: The Docker image registry.
            release: The release configuration.
            step_info: The step information.
            version_tag: The Docker image version.
        """
        super().__init__()
        self.config = ProjectConfig()
        self.project = self.config.project
        self.build = self.config.build
        self.release = self.config.release
        self.git_client = GitClient(client_root=self.project.project_root)
        self.step_info = DotMap()
        self.pre_release_num = self.build.build_num
        # Used by Docker builds
        self.registry_client = None
        self.docker_client = None
        self.image_name = ''
        self.version_tag = ''
        self.image_tag = ''

    def __getattr__(self, attr: str):
        if not hasattr(self.project, attr):
            raise AttributeError(f'No such attribute: {attr}')
        return getattr(self.step_info, attr) if getattr(self.step_info, attr) else getattr(self.project, attr)

    def _docker_init(self, *, login: bool = True) -> None:
        """Perform Docker initialization.

        Args:
            login (optional, default=True): If True, login to the Docker image registry.
            is_pre_release (optional, default=False): If this is pre-release don't add the build_num to the image_tag.

        Returns:
            Nothing.
        """
        self.registry_client = Cloud(CloudType[self.project.container_registry.type], auth=self.project.container_registry.auth, login=login)
        self.docker_client = Cloud(CloudType.local)
        registry_name_path = f'{self.project.container_registry.name}/' if login else ''
        self.image_name = f'{registry_name_path}{self.step_info.image if self.step_info.image else self.project.name}'
        self.version_tag = f'{self.image_name}:{self.project.version}'
        self.image_tag = (self.version_tag if (self.project.version_service.type == 'bumpver') else f'{self.version_tag}-{self.build.build_num}').lower()

    @override
    def _execute(self) -> None:
        """This method is called by the Action super class when this class's execute method is called."""
        class_name = type(self).__name__.lower().removeprefix('pre').removesuffix('step')
        action_method = f'{class_name}_{self.step_info.type}'
        if not hasattr(self, action_method):
            print(f'No method defined for action type {self.step_info.type} on {class_name}', file=stderr)
            sys_exit(1)
        getattr(self, action_method)()

    helm_chart_root = property(lambda s: Path(s.chart_root), doc='A read-only property which returns the helm chart root.')
    helm_package = property(lambda s: s.project.artifacts_dir / (s.pkg_name.lower() + '.tgz'), doc='A read-only property which returns the helm chart package name.')
    pkg_name = property(lambda s: f'{s.step_info.pkg_name if s.step_info.pkg_name else s.project.name}-{s.project.version}',
                        doc='A read-only property which returns the release package name.')

    @property
    def helm_args(self) -> dict:
        """A read-only property which returns the Helm command arguments."""
        helm_args = self.step_info.helm_args if self.step_info.helm_args else {}
        if self.step_info.values_files:
            helm_args['values'] = ','.join([str(self.project.artifacts_dir / v) for v in self.step_info.values_files])
        if self.step_info.helm_variables:
            helm_args['set'] = ','.join([f'{k}={v}' for (k, v) in self.step_info.helm_variables.items()])
        return helm_args

    @property
    def helm_repo(self) -> DotMap:
        """A read-only property which returns the Helm repo name with a randomized suffix."""
        if (self.project.helm_repository.type != 'oci') and self.project.helm_repository.url and not self.project.helm_repository.name:
            repo_name = f'vjer-{randint(0, 100)}'
            helm('repo', 'add', repo_name, self.chart_repo.url)
            helm('repo', 'update')
            self.project.helm_repository.name = repo_name
        return self.project.helm_repository

    def commit_files(self, message: str, branch: str, *files, file_updater: Optional[Callable] = None) -> None:
        """Checkin files during to the source repository."""
        self.git_client.client.add_remote_ref(remote_ref := REMOTE_REF, self.git_client.CI_REMOTE_REF, exists_ok=True)
        git('fetch', all=True, syscmd_args={'ignore_stderr': True})
        git('remote', 'update', remote_ref, prune=True, syscmd_args={'ignore_stderr': True})
        git('checkout', '-B', branch, '--track', f'{remote_ref}/{branch}', syscmd_args={'ignore_stderr': True})
        git('pull', remote_ref, branch, syscmd_args={'ignore_stderr': True})
        if file_updater:
            file_updater()
        self.git_client.client.add_files(*files)
        self.git_client.client.checkin_files(message, remote=remote_ref, all_branches=False)

    def copy_artifact(self, src: PathName, dest: Optional[str] = None, /) -> None:
        """Helper method to copy artifacts to the location expected by the continuous integration tool.

        Args:
            src: The source of the artifact to copy.
            dest (optional, default=None): The directory to which to copy the artifact.
                                            If not absolute, this will be a subdirectory of the project archive directory.

        Returns:
            Nothing.
        """
        use_dest = self.project.artifacts_dir
        if dest:
            if Path(dest).is_absolute():
                use_dest = Path(dest)
            else:
                use_dest /= dest
        self.log_message(f'Copying "{src}" to "{use_dest}"')
        if not Path(src).is_dir():
            copy(src, use_dest)
            return

        if WIN32:
            try:
                syscmd('robocopy', str(src), str(use_dest), '/S', '/DCOPY:D', '/COPY:D', '/MT', '/R:3', '/W:30', '/TBD', '/NP', '/NFL', show_stdout=True, ignore_stderr=True, append_stderr=True)
            except CMDError as err:
                if err.vars['returncode'] != 1:
                    raise
            return

        copytree(str(src), str(use_dest), dirs_exist_ok=True)

    def flit_build(self) -> None:
        """Run a Python flit build."""
        flit_builder(Path('pyproject.toml'))
        self.copy_artifact('dist')

    def helm_build(self) -> None:
        """Build a Helm chart."""
        helm('package', self.helm_chart_root)
        self.copy_artifact(self.helm_package.name)

    def setuptools_build(self) -> None:
        """Run a Python setuptools build."""
        flit_builder(Path('pyproject.toml'))
        self.copy_artifact('dist')

    def tag_images(self, source_tag: str, tags: list[str]) -> None:
        """Tag Docker images.

        Args:
            source_Tag: The tag of the existing image to which to add the new tags.
            tags: The list of tags to add.

        Returns:
            Nothing.
        """
        (image := self.registry_client.get_image(source_tag.split('/')[-1])).pull()
        for tag in tags:
            (repo, candidate_tag) = tag.split(':', 1) if (':' in tag) else ('', tag)
            sanitized_tag = sanitize_tag(candidate_tag)
            final_tag = f'{repo}:{sanitized_tag}' if (':' in tag) else sanitized_tag
            self.log_message(f'Tagging image: {final_tag}')
            image.tag(final_tag)
            image.push()

    def tag_source(self, tag: str, label: Optional[str] = None) -> None:
        """Tag the source in Git.

        Args:
            tag: The label which which to tag the source.
            label (optional, default=None): The annotation to apply to the tag.

        Returns:
            Nothing.
        """
        self.log_message('Removing local tags and adding remote')
        for local_tag in [t.strip() for t in git('tag', list=True)]:
            git('tag', local_tag, delete=True)
        self.git_client.client.add_remote_ref(REMOTE_REF, self.git_client.CI_REMOTE_REF, exists_ok=True)
        self.log_message(f'Tagging the source with {tag}')
        self.git_client.client.add_label(tag, label, exists_ok=True)
        self.git_client.client.checkin_files('Automated pipeline tag check-in [skip ci]', remote=REMOTE_REF, tags=True, all_branches=False)

    def update_version_files(self, *, reset: bool = False) -> None:  # pylint: disable=too-many-branches
        """Updates the version files for the project.

        Args:
            reset (optional, default=False): If True, reset the files.

        Returns:
            Nothing.
        """
        verb = 'Resetting' if reset else 'Updating'

        if not self.step_info.version_files:
            self.step_info.version_files = []

        match self.step_info.type:
            case 'helm':
                prefix = self.helm_chart_root
            case 'flit' | 'setuptools':
                prefix = Path(self.step_info.module if self.step_info.module else self.project.name)
            case _:
                prefix = Path('.')

        if self.step_info.type in DEFAULT_VERSION_FILES:
            self.step_info.version_files += [prefix / v for v in DEFAULT_VERSION_FILES[self.step_info.type] if (prefix / v).exists()]

        if not self.step_info.version_files:
            return

        self.log_message(f'{verb} version files', True)
        for file_name in self.step_info.version_files:
            file_path = Path(file_name)
            msg = f'{verb}: {file_path}'
            file_orig = Path(str(file_path) + '.orig')
            if reset:
                if file_orig.exists():
                    self.log_message(msg)
                    if file_path.exists():
                        file_path.unlink()
                    file_orig.rename(file_path)
            else:
                self.log_message(msg)
                file_path.chmod(file_path.stat().st_mode | S_IWUSR)
                copyfile(file_path, file_orig)
                file_expander(file_orig, file_path, var_props=(self.project, self.build, self.step_info))
                if file_path.name == HELM_CHART_FILE:
                    with open(file_path, encoding=DEFAULT_ENCODING) as yaml_stream:
                        helm_info = yaml_load(yaml_stream)
                    helm_info['version'] = self.project.version
                    if self.step_info.set_app_version:
                        helm_info['appVersion'] = self.project.version
                    with open(file_path, 'w', encoding=DEFAULT_ENCODING) as yaml_stream:
                        yaml_dump(helm_info, yaml_stream)


class VjerAction:  # pylint: disable=too-few-public-methods
    """This is a base class to build CI/CD support scripts using the BatCave automation module Action class."""
    def __init__(self, action_type: str, action_step_class: VjerStep):
        """
        Args:
            action_type: The action type.
            action_step_class: The Class to use for the action.

        Attributes:
            action_step_class: The value of the action_step_class argument.
            action_type: The value of the action_type argument.
            config: The project configuration.
        """
        self.config = ProjectConfig()
        self.action_type = action_type
        self.action_step_class = action_step_class

    def execute(self) -> None:
        """Run the action."""
        for (category, info) in (yaml_to_dotmap(TOOL_REPORT) if TOOL_REPORT.exists() else tool_reporter()).items():
            VjerStep().log_message(category.replace('_', ' ').title(), True)
            for (name, data) in info.items():
                VjerStep().log_message(f'  {name}: {data}')
        if not hasattr(action_def := getattr(self.config, self.action_type), 'steps'):
            return

        is_first_step = True
        for step in [DotMap(s) for s in action_def.steps]:
            step.is_first_step = is_first_step
            VjerStep().log_message(f'Executing {self.action_type} step: {step.type if (not step.name) else step.name}', True)
            (executor := cast(Callable, self.action_step_class)()).step_info = step
            executor.execute()
            is_first_step = False


def sanitize_tag(tag: str, replacement_char: str = '-') -> str:
    """Sanitize a Docker tag by replacing invalid characters with a specified valid character.

        Args:
            tag: The Docker tag to sanitize.
            replacement_char (optional, default=_): The character to replace invalid characters with.

        Returns:
            The sanitized Docker tag.
    """
    valid_pattern = re_compile(r'^[a-zA-Z0-9][a-zA-Z0-9.-]{0,127}$')
    sanitized_tag = re_sub(r'[^a-zA-Z0-9.-]', replacement_char, tag)
    if sanitized_tag[0] in ['.', '-']:
        sanitized_tag = replacement_char + sanitized_tag[1:]
    sanitized_tag = sanitized_tag[:128]
    if not valid_pattern.match(sanitized_tag):
        raise ValueError(f"The sanitized tag '{sanitized_tag}' is still not valid according to Docker's specifications.")
    return sanitized_tag

# cSpell:ignore syscmd cloudmgr platarch checkin
